import pytest

from lionsc.yaml_parser import YamlParser
from lionsc.lmsg import LMsg, MsgField, _used_ids, _used_names
from lionsc.errors import *


def reset(func):
    def wrapper(*args, **kwargs):
        _used_ids.clear()
        _used_names.clear()
        result = func(*args, **kwargs)
        return result

    return wrapper


@reset
def test_invalid_dir():
    with pytest.raises(MsgFilesNotFoundError):
        YamlParser("tests/test_files")


@reset
def test_single_lmsg_file1():
    parser = YamlParser("tests/test_files/single_lmsg_file1")

    results = []
    for r in parser.parse_file():
        results.append(r)

    assert len(results) == 1

    _used_ids.clear()
    _used_names.clear()

    f1 = MsgField(
        parent_msg_name="accelerometer", name="acc_x", type="float", size=4, start=0
    )
    f2 = MsgField(
        parent_msg_name="accelerometer", name="acc_y", type="float", size=4, start=4
    )
    f3 = MsgField(
        parent_msg_name="accelerometer", name="acc_z", type="float", size=4, start=8
    )

    answer = LMsg(id=1, name="accelerometer", period=1000, fields=[f1, f2, f3])

    filename, r = results[0]

    assert filename == "single"
    assert r[0] == answer


@reset
def test_single_lmsg_file2():
    parser = YamlParser("tests/test_files/single_lmsg_file2")

    _, r = next(parser.parse_file())
    _used_ids.clear()
    _used_names.clear()

    # ************ msg 1 ***************
    f1 = MsgField(
        parent_msg_name="accelerometer", name="acc_x", type="float", size=4, start=0
    )
    f2 = MsgField(
        parent_msg_name="accelerometer", name="acc_y", type="float", size=4, start=4
    )
    f3 = MsgField(
        parent_msg_name="accelerometer", name="acc_z", type="float", size=4, start=8
    )
    answer = LMsg(id=1, name="accelerometer", period=1000, fields=[f1, f2, f3])
    assert r[0] == answer

    # ************ msg 2 ***************
    f1 = MsgField(
        parent_msg_name="microphone",
        name="sound_level",
        type="int16_t",
        size=2,
        start=0,
    )
    f2 = MsgField(
        parent_msg_name="microphone",
        name="message",
        type="string",
        size=100,
        start=2,
    )
    answer = LMsg(id=2, name="microphone", period=0, fields=[f1, f2])
    assert r[1] == answer

    # ************ msg 3 ***************
    answer = LMsg(id=3, name="ping", period=1000, fields=[])
    assert r[2] == answer


@reset
def test_multiple_lmsg_files():
    parser = YamlParser("tests/test_files/multiple_lmsg_files")

    for filename, r in parser.parse_file():
        _used_ids.clear()
        _used_names.clear()

        if filename == "a":
            f1 = MsgField(
                parent_msg_name="accelerometer",
                name="acc_x",
                type="float",
                size=4,
                start=0,
            )
            f2 = MsgField(
                parent_msg_name="accelerometer",
                name="acc_y",
                type="float",
                size=4,
                start=4,
            )
            f3 = MsgField(
                parent_msg_name="accelerometer",
                name="acc_z",
                type="float",
                size=4,
                start=8,
            )
            answer = LMsg(id=1, name="accelerometer", period=1000, fields=[f1, f2, f3])
            assert r[0] == answer

            answer = LMsg(id=3, name="ping", period=1000, fields=[])
            assert r[1] == answer

        elif filename == "b":
            f1 = MsgField(
                parent_msg_name="microphone",
                name="sound_level",
                type="int16_t",
                size=2,
                start=0,
            )
            f2 = MsgField(
                parent_msg_name="microphone",
                name="message",
                type="string",
                size=100,
                start=2,
            )
            answer = LMsg(id=2, name="microphone", period=0, fields=[f1, f2])
            assert r[0] == answer


@reset
def test_invalid_lmsg_file():
    parser = YamlParser("tests/test_files/invalid_lmsg_file")

    with pytest.raises(MissingFieldError):
        for filename, r in parser.parse_file():
            _used_ids.clear()
            _used_names.clear()

            if filename == "valid":
                answer = LMsg(id=3, name="ping", period=1000, fields=[])
                assert r[0] == answer


@reset
def test_invalid_field_size():
    parser = YamlParser("tests/test_files/invalid_field_size")

    with pytest.raises(InvalidTypeSizeError):
        for filename, r in parser.parse_file():
            _used_ids.clear()
            _used_names.clear()

    assert True



"""
Module Name: yaml_parser.py
Author: Diogo Ferreira (ItsNotSoftware)
Date: May 8, 2024

Description:
    This module contains the YAML parser class for parsing lmsg.yaml files into python objects.

License:
    Copyright (c) 2024 Diogo Ferreira. All rights reserved.
    This code is licensed under the MIT License.
"""

import yaml

from lionsc.lmsg import LMsg, MsgField
from typing import Generator
from lionsc.errors import *
import os

from colorama import Fore, Style


class YamlParser:
    """
    Class to parse lmsg.yaml files into python objects

    Attributes:
        file_data (dict[str, dict]): Dictionary containing the filename and the data for each file

    Methods:
        validate_type_size(msg_name: str, field_name, type: str, size: int): Validate the size of the field based on the type
        get_file_data(msg_files_dir: str) -> dict[str, dict]: Load the lmsg.yaml files from the directory msg_files_directory
        yamlMsg_to_LMsg(msg_name: str, msg_data: dict) -> LMsg: Convert a YAML message to a LMsg object
        parse_file() -> Generator[tuple[str, list[LMsg]], None, None]: Parse the lmsg.yaml files and yields information about one file at a time
    """

    def __init__(self, msg_files_dir: str):
        """
        Initialize the YamlParser object

        Args:
            msg_files_dir (str): Directory containing the lmsg.yaml files
        """
        self.file_data = self.get_file_data(msg_files_dir)

    @staticmethod
    def validate_type_size(msg_name: str, field_name, type: str, size: int):
        """
        Validate the size of the field based on the type

        Args:
            msg_name (str): Name of the message
            field_name (str): Name of the field
            type (str): Type of the field
            size (int): Size of the field

        Raises:
            InvalidTypeSizeError: If the size is invalid for the given type
        """

        if type == "string" and size <= 1:
            raise InvalidTypeSizeError(msg_name, field_name, type, size)

        if type in ["bool", "uint8_t", "int8_t"] and size != 1:
            raise InvalidTypeSizeError(msg_name, field_name, type, size)

        if type in ["uint16_t", "int16_t"] and size != 2:
            raise InvalidTypeSizeError(msg_name, field_name, type, size)

        if type in ["uint32_t", "int32_t", "float"] and size != 4:
            raise InvalidTypeSizeError(msg_name, field_name, type, size)

        if type in ["uint64_t", "int64_t", "double"] and size != 8:
            raise InvalidTypeSizeError(msg_name, field_name, type, size)

    @staticmethod
    def get_file_data(msg_files_dir: str) -> dict[str, dict]:
        """
        Load the lmsg.yaml files from the directory msg_files_directory

        Args:
            msg_files_dir (str): Directory containing the lmsg.yaml files

        Raises:
            MsgFilesNotFoundError: If no lmsg.yaml files are found in the directory

        Returns:
            dict[str, dict]: Dictionary containing the filename and the data for each file
        """
        file_data = {}

        # Get the list of files in the directory
        try:
            subdirs = os.listdir(msg_files_dir)
        except FileNotFoundError:
            raise MsgFilesNotFoundError(msg_files_dir)

        for file in subdirs:
            if file.endswith(".lmsg.yaml"):
                # get filename without extension
                filename = file.split("/")[-1].split(".")[0]

                file_data[filename] = yaml.safe_load(open(f"{msg_files_dir}/{file}"))

        # Check if any lmsg.yaml files are found
        if not file_data:
            raise MsgFilesNotFoundError(msg_files_dir)

        # Check if any of the data files are empty
        for key in file_data:
            if file_data[key] is None:
                print(
                    Fore.YELLOW
                    + "Warning: Empty file found: "
                    + msg_files_dir
                    + "/"
                    + key
                    + ".lmsg.yaml\n"
                )

        return file_data

    @staticmethod
    def yamlMsg_to_LMsg(msg_name: str, msg_data: dict) -> LMsg:
        """
        Convert a YAML message to a LMsg object

        Args:
            msg_name (str): Name of the message
            msg_data (dict): Dictionary containing the message data

        Raises:
            ValueError: If a required key is missing in the message data

        Returns:
            LMsg: LMsg object

        """

        # Extract the required fields from the message data
        try:
            id = msg_data["id"]
            name = msg_name
            period = msg_data["period"]
        except KeyError as e:
            key = e.args[0]
            raise MissingFieldError(key, msg_name)

        fields = []
        start = 0  # Start index for the first field

        # If the message has fields iterate over them
        if msg_data.get("fields") is not None:
            for field_name_, field_data in msg_data["fields"].items():

                # Extract the field data
                try:
                    field_name = field_name_
                    field_type = field_data["type"]
                    field_size = field_data["size"]

                    YamlParser.validate_type_size(
                        msg_name, field_name, field_type, field_size
                    )

                except KeyError as e:
                    key = e.args[0]
                    raise MissingFieldError(key, msg_name + "/" + field_name_)

                fields.append(
                    MsgField(
                        parent_msg_name=name,
                        name=field_name,
                        type=field_type,
                        size=field_size,
                        start=start,
                    )
                )
                start += field_size  # Update the start index for the next field

        return LMsg(id=id, name=msg_name, period=period, fields=fields)

    def parse_file(self) -> Generator[tuple[str, list[LMsg]], None, None]:
        """
        Parse the lmsg.yaml files and yields information about one file at a time

        Yields:
            tuple[str, list[LMsg]]: Tuple containing the filename and the LMsg object for each file
        """

        # Iterate over each file
        for filename, data in self.file_data.items():
            # Ignore empty files
            if not data:
                continue

            lmsg_list = []

            # Iterate over each message in the file
            for msg_name, msg_data in data.items():
                lmsg_list.append(self.yamlMsg_to_LMsg(msg_name, msg_data))

            # Return the filename and the LMsg object for each file one at a time
            yield filename, lmsg_list



"""
Module Name: errors.py
Author: Diogo Ferreira (ItsNotSoftware)
Date: May 8, 2024

Description:
    This module contains the custom exceptions used by the LIONS compiler.
    This exceptions are used for error messages in case of invalid message files. 

License:
    Copyright (c) 2024 Diogo Ferreira. All rights reserved.
    This code is licensed under the MIT License.
"""

from colorama import Fore, Style


class OutOfBoundsIdError(Exception):
    """Exception raised when a message ID is out of bounds."""

    def __init__(self, msg_name, id):
        message = f'Error in message "{msg_name}": ID "{id}" is out of bounds. Valid range: 0-255.'
        super().__init__(Fore.RED + message)


class MissingFieldError(Exception):
    """Exception raised when a required field is missing."""

    def __init__(self, field, message_name):
        message = (
            f'Error in message "{message_name}": Missing required field "{field}".'
        )
        super().__init__(Fore.RED + message)


class DuplicateIdError(Exception):
    """Exception raised when a message ID is duplicated."""

    def __init__(self, name, id, sugested_id):
        message = f'Error in message "{name}": Duplicate ID "{id}". Suggested alternative: "{sugested_id}".'
        super().__init__(Fore.RED + message)


class DuplicateMsgNameError(Exception):
    """Exception raised when a message name is duplicated."""

    def __init__(self, name):
        message = f'Error: Duplicate message name "{name}".'
        super().__init__(Fore.RED + message)


class InvalidTypeSizeError(Exception):
    """Exception raised when a field has an invalid size for its type."""

    def __init__(self, msg_name, field_name, type, size):
        expected_sizes = {
            "bool": 1,
            "int8_t": 1,
            "uint8_t": 1,
            "int16_t": 2,
            "uint16_t": 2,
            "int32_t": 4,
            "uint32_t": 4,
            "int64_t": 8,
            "uint64_t": 8,
            "float": 4,
            "double": 8,
            "string": ">1",
        }
        message = f'Error in message "{msg_name}", field "{field_name}": Invalid size "{size}" for type "{type}". Expected size: {expected_sizes[type]}.'
        super().__init__(Fore.RED + message)


class InvalidTypeError(Exception):
    """Exception raised when a field has an invalid type."""

    def __init__(self, msg_name, field_name, type):
        message = f'Error in message "{msg_name}", field "{field_name}": Invalid type "{type}".'
        super().__init__(Fore.RED + message)


class PayloadOverflowError(Exception):
    """Exception raised when the payload size exceeds the maximum size."""

    def __init__(self, msg_name, size):
        message = f'Error in message "{msg_name}": Payload size "{size}" exceeds maximum size of 244 bytes. Please reduce the size of the message fields'
        super().__init__(Fore.RED + message)


class MsgFilesNotFoundError(Exception):
    """Exception raised when the message file is not found."""

    def __init__(self, dir):
        message = f'Error: No ".lmsg.yaml" files found in the directory "{dir}".'
        super().__init__(Fore.RED + message)


class InvalidTargetLanguageError(Exception):
    """Exception raised when the target language is invalid."""

    def __init__(self, target_language):
        message = f'Error: Invalid target language "{target_language}".'
        super().__init__(Fore.RED + message)



"""
Module Name: main.py
Author: Diogo Ferreira (ItsNotSoftware)
Date: May 8, 2024

Description:
    This module is the main entry point for the LIONS compiler. It is responsible for
    parsing the command line arguments, reading the message files, and generating the
    corresponding C++ message files.

License:
    Copyright (c) 2024 Diogo Ferreira. All rights reserved.
    This code is licensed under the MIT License.
"""

import sys
import colorama
from lionsc.code_generation.cpp.cpp_generator import CppGenerator
from lionsc.code_generation.js.js_generator import JsGenerator
from lionsc.code_generation.ts.ts_generator import TsGenerator
from lionsc.code_generation.python.python_generator import PythonGenerator
from lionsc.yaml_parser import YamlParser
from colorama import Fore, Style
from lionsc.errors import InvalidTargetLanguageError


def invalid_language(_):
    """Raise an InvalidTargetLanguageError if the target language is invalid."""
    raise InvalidTargetLanguageError(sys.argv[3])


def print_title():
    """Function to print title of the program in ASCII art"""

    print(Fore.GREEN)

    print(
        """
██╗     ██╗ ██████╗ ███╗   ██╗███████╗
██║     ██║██╔═══██╗████╗  ██║██╔════╝
██║     ██║██║   ██║██╔██╗ ██║███████╗
██║     ██║██║   ██║██║╚██╗██║╚════██║
███████╗██║╚██████╔╝██║ ╚████║███████║
╚══════╝╚═╝ ╚═════╝ ╚═╝  ╚═══╝╚══════╝
"""
    )
    print(Style.RESET_ALL + "\nInitiating compilation of messages...\n")


def print_success_message():
    """Function to print success message after all messages are compiled"""

    success_message = "All messages compiled successfully!"
    print(Fore.GREEN + "+" + "-" * (len(success_message) + 2) + "+" + Style.RESET_ALL)
    print(Fore.GREEN + f"| {success_message} |" + Style.RESET_ALL)
    print(Fore.GREEN + "+" + "-" * (len(success_message) + 2) + "+" + Style.RESET_ALL)


def main():
    colorama.init(autoreset=True)

    # Check if the number of arguments is correct
    if len(sys.argv) != 4:
        print("Usage: lionsc <msg_files_dir> <output_dir> <target_language>")

        # targert language help
        print(
            """
    Target Language:
        cpp - C++
        js - JavaScript
        ts - TypeScript
        py - Python
        """
        )

        sys.exit(1)

    msg_files_dir = sys.argv[1]
    output_dir = sys.argv[2]

    print_title()

    # Initialize the parser and generator objects
    parser = YamlParser(msg_files_dir)

    # Initialize the code generator based on the target language
    code_generator = {
        "cpp": CppGenerator,
        "js": JsGenerator,
        "ts": TsGenerator,
        "py": PythonGenerator,
    }.get(sys.argv[3], invalid_language)(output_dir)

    # Parse the message files and generate the corresponding C++ files
    for filename, msgs in parser.parse_file():
        code_generator.generate_msg_files(filename, msgs)

    print_success_message()



from jinja2 import Environment, FileSystemLoader
from lionsc.lmsg import LMsg
import time
from colorama import Fore, Style
import os


def print_generation_status(output_dir, file_name):
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"{timestamp} - Successfully generated: {output_dir}/{file_name}")


class PythonGenerator:
    def __init__(self, output_dir: str):
        """
        Constructor for the PythonGenerator class

        Args:
            output_dir (str): The output directory where the generated files will be saved
        """

        self.output_dir = output_dir

        template_dir = os.path.dirname(__file__) + "/templates"
        self.env = Environment(loader=FileSystemLoader(template_dir))

        # Load the templates
        self.module_py_template = self.env.get_template("lions.py.jinja")
        self.class_py_template = self.env.get_template("msg.py.jinja")

        self.generate_lions_py()

    @staticmethod
    def convert_to_py_types(lmsgs: list[LMsg]) -> list[LMsg]:
        """
        Convert the message types to TypeScript types

        Args:
            lmsgs (list[LMsg]): The list of messages to convert

        Returns:
            list[LMsg]: The list of messages with converted types
        """

        cpp_to_py_bufs = {
            "int8_t": "b",
            "uint8_t": "B",
            "int16_t": "<h",
            "uint16_t": "<H",
            "int32_t": "<i",
            "uint32_t": "<I",
            "int64_t": "<q",
            "uint64_t": "<Q",
            "float": "<f",
            "double": "<d",
            "string": "s",
            "bool": "?",
        }

        cpp_to_py_types = {
            "int8_t": "int",
            "uint8_t": "int",
            "int16_t": "int",
            "uint16_t": "int",
            "int32_t": "int",
            "uint32_t": "int",
            "int64_t": "int",
            "uint64_t": "int",
            "float": "float",
            "double": "float",
            "string": "str",
            "bool": "bool",
        }

        for msg in lmsgs:
            for field in msg.fields:
                field.buff_type = cpp_to_py_bufs[field.type]
                field.type = cpp_to_py_types[field.type]

        return lmsgs

    def generate_lions_py(self):
        """Generate the lions.py file"""

        with open(f"{self.output_dir}/lions.py", "w") as f:
            f.write(self.module_py_template.render())
            print_generation_status(self.output_dir, "lions.py")

    def generate_msg_files(self, filename: str, msgs: list[LMsg]):
        """
        Generate the .py message files for the given message file (yaml)

        Args:
            filename (str): The filename of the message files
            msgs (list[LMsg]): The list of messages to generate
        """
        msgs = self.convert_to_py_types(msgs)

        # Dictionary to pass to the jinja templates
        jinja_dict = {"filename": filename, "msgs": msgs}

        file = f"{self.output_dir}/{filename}"
        with open(f"{file}_lmsg.py", "w") as f:
            f.write(self.class_py_template.render(jinja_dict))
            print_generation_status(self.output_dir, f"{filename}.py")
